/* *************************************************

Algorithm LB-treedec
input  A graph G =(V,E)
output A clique tree T of G

**************************************************** */

#include <LEDA/graph/graph.h>
#include <LEDA/graph/node_array.h>
#include <LEDA/core/slist.h>
#include <LEDA/core/list.h>
#include <LEDA/core/array.h>
#include <LEDA/core/string.h>
#include <LEDA/graphics/graphwin.h>
#include <iostream>
using namespace leda;
using namespace std;

void tree(graph &G,GRAPH<list<node>,int> &T);
void ope(graph &G,GRAPH<list<node>,int> &T,node v);
void c_component(graph &G,node_array<int> &node_ord);
void N_G(graph &G);
int existU_A();
void existX(graph &G,GRAPH<list<node>,int> &T,list<node> &X);
void replace(graph &G,GRAPH<list<node>,int> &T,node v);
void rep(GRAPH<list<node>,int> &T,list<node> &X_1,list<node> &X_2,node v);
void f_edge(GRAPH<list<node>,int> &T,list<node> &kl,node v);

list< list<node> > U_A; //U_A:minimal separator
list<node> C,L,S;       //C:C_G(N_H[v])  L:C in C_G S:N_G(C)

int main()
{
  graph G;
  G.read("graph1.gw");
  GRAPH<list<node>,int> T;
  G.make_undirected();
  T.make_undirected();

  node v;
  forall_nodes(v,G)
    cout << v << endl;
  tree(G,T);
  node w;
  cout << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl;
  forall_nodes(w,T){
    cout << T[w] << endl;
    forall(v,T[w])
      cout << v << " " ;
    cout << endl << endl;
  }
  T.write("out.gw");
}

void tree(graph &G,GRAPH<list<node>,int> &T)
{
  U_A.clear();
  node w = T.new_node();
  node v;
  forall_nodes(v,G)
    T[w].append(v);
  forall_nodes(v,G){
    //  cout << "#################################" << endl;
    ope(G,T,v);
    // GraphWin gw(T,"LEDA Graph Editor");
    // gw.display(window::center,window::center);
    // gw.edit();
     node v,w;
     //   forall_nodes(v,T){
     //  forall(w,T[v])
     //	 cout << w << endl;
     // cout << "************* " << endl;
     // }
  }
}
void ope(graph &G,GRAPH<list<node>,int> &T,node v)
{
  C.clear();
  L.clear();
  S.clear();

  //C_G(N_H[i])
  node_array<int> C_G(G,1);
  node w;
  forall_adj_nodes(w,v)
    C_G[w]=0;
  list_item it,its;
  forall_items(it,U_A)
    if(U_A[it].search(v) != nil)
      forall_items(its,U_A[it])
	C_G[ U_A[it][its] ] = 0;
  C_G[v]=0;

  //C_G(N_H[v])$B$r(Blist C$B$KJ]B8$9$k(B
  forall_nodes(w,G)
    if(C_G[w]==1)
      C.append(w);
  while( !C.empty() ){
    c_component(G,C_G);
    N_G(G);
    if(!existU_A())
      U_A.append(S);

    // bag X in T containing v and S and C
    list<node> X;;
    forall(w,S)
      X.append(w);
    if(X.search(v) == nil)
      X.append(v);
    forall(w,L)
      if(X.search(w) == nil)
	X.append(w);
    existX(G,T,X);
  }

}

//connected component C in C_G(N_H[v])
void c_component(graph &G,node_array<int> &node_ord)
{
  list_item it = C.first();

  node v = C[it];
  node_ord[v] = 0;
  L.append(v);
  C.erase(it);

  node w;
  forall_adj_nodes(w,v)
    if( node_ord[w]==1 )
      c_component(G,node_ord);
}

//N_G(C)$B$r(BS$B$KJ]B8$9$k(B
void N_G(graph &G)
{
  node v,w;
  forall(v,L){
    forall_adj_nodes(w,v){
      if(S.search(w)==nil)
	S.append(w);
    }
  }
  forall(v,L)
    S.remove(v);
}


int existU_A()
{

  list_item S_it = S.first();
  list_item U_it = U_A.first();
 
  list<node> K;
  node v;
  forall(K,U_A){
    int i=0;
    forall(v,S)
      if(K.search(v) != nil)
	i++;
    if(i==S.size())
      return 1;
  }
  return 0;
}

void existX(graph &G,GRAPH<list<node>,int> &T,list<node> &X)
{

  node v,w;
  forall_nodes(v,T){
    int i=0;
    forall(w,X)
      if(T[v].search(w) != nil)
	i++;
    if(i==X.size())
      replace(G,T,v);
  }
}


void replace(graph &G,GRAPH<list<node>,int> &T,node v)
{
  list<node>X_1;
  list<node>X_2;
  X_1.clear();
  X_2.clear();

  //X_1$B:n@.(B
  node_array<int> n1(G,0);
  node w;
  forall(w,T[v])
    n1[w]=1;

  forall(w,L)
    if(n1[w] == 1)
      n1[w] = 2;

  forall(w,S)
    X_1.append(w);
  forall_nodes(w,G)
    if(n1[w] == 2 && X_1.search(w)==nil)
      X_1.append(w);


  //X_2$B:n@.(B
  node_array<int> n2(G,1);
  forall(w,L)
    n2[w] = 0;
  forall(w,T[v])
    if(n2[w]==1)
      X_2.append(w);


  list<node> kl;
  kl.clear();
  if(U_A.size() > 1)
    f_edge(T,kl,v);
  else
    if(U_A.size() == 1)
      kl = U_A.head();

  node m;
  int count =0;
  if(kl.size() != 0)
    forall(m,kl){
      if(X_1.search(m) != nil)
	count++;
    }
  if(count == kl.size() )
    rep(T,X_1,X_2,v);
  else
    rep(T,X_2,X_1,v);
}

void rep(GRAPH<list<node>,int> &T,list<node> &X_1,list<node> &X_2,node v)
{

  T[v].clear();
  node w1;
  forall(w1,X_1)
    T[v].append(w1);
  
  node v1 = T.new_node();
  T[v1].clear();
  node w2;
  forall(w2,X_2)
    T[v1].append(w2);
  
  edge e = T.new_edge(v,v1);
}

void f_edge(GRAPH<list<node>,int> &T,list<node> &kl,node v)
{
  list_item it = U_A.first();
  node w;
  while(it != nil){
    int count = 0;
    forall(w,U_A[it])
      if(T[v].search(w) != nil)
	count++;
    if(count == U_A[it].size() ){
      kl = U_A[it];
      return ;
    }
    it = U_A.succ(it);
  }
}
